/* Main wrapper for maximal clique enumerator
 *
 * Author: Hoony Park
 * Last modified: 
 *
 * Copyright 2006-2007
 * Computer Science and Mathematics Division, Oak Ridge National Laboratory
 *
 */

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>

#if defined(USE_PTHREADS)
#include <pthread.h>
#endif

#include "graph.h"
#include "clique.h"
#include "msg.h"
#include "stats.h"

static char *graph_file = NULL;			/* input graph file name */

/* Maximum vertices for an adjacency matrix */
static long matrix_cutoff = 100000L;		/* 1.25 GB matrix (nv*nv/8) */

/* Maximum edges in graph for a hash table. */
static long hash_cutoff = 50000000L;		/* 1.2 GB table (ne*xf*4) */

void print_usage(void)
{
  fprintf(stderr, "Usage: pDFS [options] -f <FileName>\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -f <FileName> Input graph file\n");
  fprintf(stderr, "  -o <FileName> Output file to store cliques\n");
  fprintf(stderr, "  -a <#verts>   Maximum number of vertices in the adjacency matrix\n");
  fprintf(stderr, "                (optional, default is %d)\n",matrix_cutoff);
  fprintf(stderr, "  -l <#value>   Lower bound of clique size\n");
  fprintf(stderr, "                (optional, has to be >= 2, default = 2.)\n");
  fprintf(stderr, "  -u <#value>   Upper bound of clique size\n");
  fprintf(stderr, "                (optional, has to be >= 3, default = maximum clique size.)\n");
  fprintf(stderr, "  -n <#threads> (optional, has to be >= 1, default = 1\n");
  fprintf(stderr, "  -b 0|1        Turn load balancing on (1, default) or off (0).\n");
  fprintf(stderr, "  -p            Print per-process/per-thread run statistics.\n");
  fprintf(stderr, "  -s <time>     Sample interval for time-varying statistics.\n");
  fprintf(stderr, "                Default is 0 (no time-varying statistics).\n");
  fprintf(stderr, "  -t <#edges>   Maximum number of edges in the hash table\n");
  fprintf(stderr, "                (optional, default is %ld)\n",hash_cutoff);
  fprintf(stderr, "  -e <FileName> Output file for load balance events\n");
  fprintf(stderr, "  -c            Print table of clique counts by size.\n");
}


void clean_up()
{
    int i;
    if ( OUTPUT_SET ) {
	for (i=0; i<num_threads; i++) fclose(CG.CliqueFP[i]);
    }
}


void parse_param(int argc, char **argv)
{
    int i,c, errflag = 0;
    char *stem;
    char name[100];

    while ( (c=getopt(argc, argv, "a:b:ce:f:l:n:o:ps:t:u:")) != -1 ) {
	switch ( c ) {
	    case 'a':
		matrix_cutoff = atol(optarg);
		break;
	    case 'b':
		dynamic_load_balancing = atoi(optarg);
		break;
	    case 'c':
		clique_breakdown = 1;
		break;
	    case 'e':
		event_file = optarg;
		break;
	    case 'f':
		graph_file = optarg;
		break;
	    case 'l':
		CG.InitCliqueSize = atoi(optarg);
		break;
	    case 'n':
		num_threads = atoi(optarg);
		break;
	    case 'o':
                if (strlen(optarg) > 90) {
                   if (mpi_rank == 0) fprintf(stderr,"Output file prefix is too long (-o).\n");
                   exit(1);
                }
		stem = optarg;
		OUTPUT_SET = 1;
		break;
	    case 'p':
		full_statistics = 1;
		break;
	    case 's':
		sample_time = atof(optarg);
		break;
	    case 't':
		hash_cutoff = atol(optarg);
		break;
	    case 'u':
		CG.EndCliqueSize = atoi(optarg);
		break;
	    default:
		errflag++;
		break;
	}
    }

    if (graph_file == NULL) errflag = 1;

    if ( errflag ) {
	if (mpi_rank == 0) print_usage();
	exit(1);
    }

    if ( OUTPUT_SET ) {
	if ( (CG.CliqueFP = (FILE **)malloc(num_threads*sizeof(FILE *))) == NULL ) {
	    fprintf(stderr, "%d: malloc error\n",mpi_rank);
	    exit(-1);
	}
	
	for (i=0; i<num_threads; i++) {
#if defined(USE_MPI) && defined(USE_PTHREADS)
            sprintf(name,"%s.%04d-%03d",stem,mpi_rank,i);
#elif defined(USE_MPI)
            sprintf(name,"%s.%04d",stem,mpi_rank);
#elif defined(USE_PTHREADS)
            sprintf(name,"%s.%03d",stem,i);
#else
            sprintf(name,"%s",stem);
#endif
	    if ((CG.CliqueFP[i] = fopen(name, "w")) == NULL) {
		fprintf(stderr, "Can't open output file \"%s\"\n",name);
		exit(-1);
	    }
#if defined(__QK_USER__)
            /* For Catamount, use a large user buffer because there is no kernel buffer. */
            if (setvbuf(CG.CliqueFP[i],malloc(1024*1024),_IOFBF,1024*1024))
              fprintf(stderr,"%d: Unable to set buffering for thread %d\n",mpi_rank,i);
#endif
	}
    }
}


int main(int argc, char  **argv)
{
    int rank = 0;
    double init_start, st;

    init_start = dclock();
    rank = msg_handler_init(&argc,&argv);
    parse_param(argc, argv);	/* parse the command line arguments */

    rstats_init();		/* initialize the run time statistics gathering */

    st = dclock();
    G = read_graph(graph_file,matrix_cutoff,hash_cutoff);
    rstats[0].read_time += dclock() - st;

    if (rank == 0) printf("%d processes with %d threads, %d vertices, %d edges, %d max degree\n",mpi_size,num_threads,G->num_vertices,G->num_edges,G->max_degree);

    /* Prepare to handle dequeue requests with a maximum size of num_threads. */
    msg_handler_start(num_threads);
    
    init_memory();
    rstats[0].init_time += dclock() - init_start;
    init_graph();

#if defined(USE_PTHREADS)
    if (num_threads == 1) {
	/* Use master thread alone. */
	clique_enumerate((void *) 0);
    } else {
	/* Use worker threads. */
	pthread_t *threads;
	pthread_attr_t attr;
	int i;

	if ( (threads = (pthread_t *)malloc(num_threads*sizeof(pthread_t))) == NULL ) {
	    fprintf(stderr,"malloc error\n");
	    exit(1);
	}

	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

	/* Start other worker threads. */
	for (i=1; i<num_threads; i++) {
	    int err = pthread_create(threads+i, &attr, clique_enumerate, (void *)i);
	    if (err != 0 ) fprintf(stderr,"ERROR: in creating pthread %d (%d)\n",i,err);
	}

	clique_enumerate((void *) 0);	/* master is thread 0 */

	/* Wait for the other worker threads to complete. */
	for (i = 1; i < num_threads; i++) {
	    pthread_join(threads[i], NULL);
	}

	free(threads);
    }
#else
    /* MPI-only or serial. */
    clique_enumerate((void *) 0);
#endif

    st = dclock();
    msg_handler_complete();	/* wait for all processes to go idle */
    rstats[0].idle_time += dclock() - st;

    rstats_end();		/* print statistics */
    msg_handler_stop();		/* MPI_Finalize */
    free_graph(G);
    clean_up();

    return 0;
}
