/* Clique Enumerator
 * Data intensive, memory intensive, computation intensive
 *
 * Author: Hoony Park
 * Date: 
 *
 * Copyright 2006-2007
 * Computer Science and Mathematics Division, Oak Ridge National Laboratory
 *
 * Last Modified:
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include "clique.h"
#include "msg.h"
#include "stats.h"

/* Thread synchronization routines. */
static void thread_init(void);
static void thread_barrier(void);
static void thread_destroy(void);

/** Global variables **/
Configure CG = { 2, 1000000, NULL };	/* global configuration */
Graph *G;
MCB *mcb_list;				/* per-thread memory control blocks */

/* User settings. */
int num_threads = 1;			/* number of threads per process */
int OUTPUT_SET = 0;			/* write cliques to a file (default=no) */
int dynamic_load_balancing = 1;		/* use work-stealing (default=yes) */
int full_statistics = 0;		/* print per-thread statistics (default=no) */

static clist_t *init_list;
static cstruct_t **CLIST;
static int *new_work = NULL;		/* thread received remote work flag */


/* Print out the i-th clique in a clique list of size k */
static void clique_out(clist_t u, clist_t k, int id)
{
  /*
  fwrite(&k,sizeof(unsigned short),1,CG.CliqueFP);
  fwrite(list,sizeof(vid_t),k-1,CG.CliqueFP);
  fwrite(&v,sizeof(vid_t),1,CG.CliqueFP);
  */

  int i;
  double st = dclock();
  for (i = 0; i < k-1; i++) {
    fprintf(CG.CliqueFP[id], "%d ", mcb_list[id].compsub[i]);
  }
  fprintf(CG.CliqueFP[id], "%d\n", u);
  rstats[id].output_time += dclock() - st;
}


/**
 *
 * "spawn_clists" is the core part of the entire program. Given a
 * (k-1)-clique list L, it create k-clique lists that L spawns.
 *
 * @param:
 *           L : input (k-1) clique list
 */
static void spawn_clists(cstruct_t *parent, int bp)
{
    clist_t i,j;
    clist_t clique_size = parent->clique_size;
    clist_t k_1         = parent->k_1;
    clist_t ne          = parent->ne;
    clist_t ce          = parent->ce;
    clist_t *source     = &parent->ce; /* same as where ce is stored. */
    clist_t k = clique_size + 1;	/* new clique size */
    clist_t *frame;
    clist_t *target;
    clist_t newne = 0;
    clist_t newce = 0;
    clist_t minod = ce;
    clist_t nod = 0;
    clist_t p, fixp, pos, s;
    clist_t count;
    double st = dclock(), et;
    MCB *CB = &mcb_list[bp];

    if ( k > CG.EndCliqueSize ) return;

    /* latent insertion of the (k-1)-th into compsub */
    if ( clique_size > 1 )
	CB->compsub[clique_size-2] = k_1;

    /* find the node with minimum non-connectivity */
    for (i=1; i<=ce && minod != 0 ; i++) {
	p = source[i]; 
	count=0;

	/* Search the list of candidates to get the non-connectivity count for p. */
	for (j=ne+1; j <= ce && count < minod; j++) {
	    if (!connected(G,p,source[j]) ) {
		count++;
		pos = j;
	    }
	}

	/* If p has the minimum non-connectivity so far, save it. */
	if ( count < minod ) {
	    fixp = p;
	    minod = count;
	    if ( i <= ne ) {
		s = pos;
	    }
	    else {
		s = i; 
		nod = 1;
	    }
	}
    }

    nod = minod + nod;

    while ( nod > 0 ) {
	clist_t sel = source[s];
	source[s] = source[ne+1];
	source[ne+1] = sel;

	CB = &mcb_list[bp];

	/* We need at most this amount of memory for i. */
	frame = MCB_next(CB, 4 + ce);

	*frame = k;
	target = (frame+4);

	newne = newce = 0;

	/* builds a new NOT */
	for (i=1; i<=ne; i++) {
	    if ( connected(G,sel,source[i]) ) {
		target[newne] = source[i];
		newne++;
	    }
	}

	newce = newne;

	/* builds a new Candidates */
	for (i=ne+2; i<=ce; i++) {
	    if ( connected(G,sel,source[i]) ) {
		target[newce] = source[i];
		newce++;
	    }
	}

	/* write out if it is a maximal clique  */
	if (newce == 0) {
	    if ( k-1 >= CG.InitCliqueSize) {
		rstats[bp].cliques++;
		if (clique_breakdown) clique_bin(k-1);
		if ( OUTPUT_SET ) clique_out(sel,k-1,bp);
	    }
	}
	    
	/* keep in the memory block only if a clique is potential candidate. */ 
	if ( newne < newce ) {
	    if ( (newce-newne + k) >= CG.InitCliqueSize ) {
		*(frame+1) = sel;
		*(frame+2) = newne;
		*(frame+3) = newce;
		MCB_advance(CB, newce+4);
	    }
	}

	ne = ne + 1;

	/* Select a candidate disconnected to the fixp */
	s = ne+1;
	if ( nod > 1 ) {
	    while ( connected(G,fixp,source[s]) ) s++;
	}

	nod--;
    }

    et = dclock();
    rstats[bp].compute_time += et - st;
    /* If accumulating run time samples, check sampling interval. */
    if (sample_time != 0) sample_check(et,bp);
}


/*
 * Create the initial list of all active vertices.
 * This is really only required if a minimum clique size is set, since
 * then some vertices will not be active.
 */
void init_graph()
{
    clist_t n = G->num_vertices;
    clist_t init_k = CG.InitCliqueSize;
    int index = 0;
    int i;
    
    /* Preprocessing the graph */
    delete_lower_degree_vertices(G, init_k);

    if (G->num_active_vertices < init_k) {
	printf("Preprocessing: there are no cliques of size %d.\n", init_k);
	exit(0);
    }

    if ( (init_list = (clist_t *)malloc(G->num_active_vertices*sizeof(clist_t))) == NULL ) {
	printf("malloc error\n");
	exit(0);
    }

    for (i=0; i<n; i++) {
	if ( G->vertex[i].active == 1 )
	    init_list[index++] = i;
    }

    thread_init();	/* initialize thread barrier */
}


/*
 * Build the initial clique list of size one (the root).
 * This is done in parallel, with each thread handling a subset of the
 * active vertices in the graph.
 */
void init_clist(int id)
{
    int gthreads = mpi_size * num_threads;	/* number of global threads */
    int gid = id + mpi_rank * num_threads;	/* this thread's global id */
    clist_t num_active = G->num_active_vertices;
    MCB *CB = &mcb_list[id];
    clist_t i,j;
    
    /* Create the initial clique list (cliques of size one), which is the set
     * of all vertices.  The initial work is divided amongst the global threads
     * by selecting a subset n/gthreads vertices. */
    for (i=gid; i<num_active; i+=gthreads) {
	clist_t *frame = MCB_next(CB, 4 + (num_active-i));
	clist_t newce = 0, newne = 0;
	clist_t root_vertex = init_list[i];
	
	/* Build NOTs from vertices before the root (init_list[0..i-1]). */
	for (j=0; j<i; j++) {
	    if ( connected(G, root_vertex, init_list[j]) )
		*(frame + 4 + newne++) = init_list[j];
	}
	
        /* Start the candidates after the NOTs. */
	newce = newne;

	/* Build Candidates from vertices past the root (init_list[i+1..n]). */
	for (j=i+1; j<num_active; j++) {
	    if ( connected(G, root_vertex, init_list[j])  )
		*(frame + 4 + newce++) = init_list[j];
	}

	/* If the candidates list is large enough, save the item. */
	if ( (newce - newne + 1) >= CG.InitCliqueSize ) {
	    *frame = 2;
	    *(frame+1) = root_vertex;
	    *(frame+2) = newne;
	    *(frame+3) = newce;
	    MCB_advance(CB, newce+4);
	}

    	/* Check for an early termination condition. */
        /* (current number of candidates) == (all remaining vertices) */
	if ( (newce-newne) == (num_active - i - 1) ) break;
    }
}


void init_memory()
{
    int gthreads = mpi_size * num_threads;	/* number of global threads */
    int i, j;

    if ( (CLIST = (cstruct_t **)malloc(num_threads*sizeof(cstruct_t *))) == NULL ) {
	fprintf(stderr,"malloc error\n");
	exit(1);
    }

    for (i=0; i<num_threads; i++) {
        int n = G->num_vertices;
	CLIST[i] = (cstruct_t *) malloc(sizeof(cstruct_t) + 2*n*sizeof(clist_t));
        if (CLIST[i] == NULL) {
	    fprintf(stderr,"malloc error\n");
	    exit(1);
	}
    }

    if ( (mcb_list = (MCB *)malloc(num_threads*sizeof(MCB))) == NULL ) {
	fprintf(stderr,"malloc error\n");
	exit(1);
    }

    for (i=0; i<num_threads; i++) {
	MCB_initialize(&mcb_list[i],G->num_vertices);
    }
}


/*
 * Randomize a list of thread ids or process ranks.
 */
void shuffle(int *list, int n)
{
    int i,j,temp;
    
    for (i=0; i < n; i++) {
	j = rand() % n;
	temp = list[j];
	list[j] = list[i];
	list[i] = temp;
    }
}


/**
 *
 * "clique_enumerate" is the main driver routine. It repeatedely pulls out a k-clique list
 * from the end of the memory block and spawn all (k+1)-cliques until memroy block becomes empty.
 *
 */
void *clique_enumerate(void *workerID)
{
    int id = (int) workerID;
    MCB *CB = &mcb_list[id];
    double root_start, enum_start;
    int *thread_list = NULL;		/* list of threads for work stealing */
    static int global_work_left = 0;
    int local_load_balancing  = dynamic_load_balancing && (num_threads > 1);
    int remote_load_balancing = dynamic_load_balancing && (mpi_size > 1);

    root_start = dclock();
    init_clist(id);
    enum_start = dclock();
    rstats[id].root_time += enum_start - root_start;

    /* Save the time=0 run time statistics sample. */
    if (sample_time != 0) sample_check(enum_start,id);

    if ( local_load_balancing ) {
	int count = 0;
        int i;
	thread_list = (int *) malloc((num_threads-1)*sizeof(int));
	if (thread_list == NULL) {
	    fprintf(stderr, "malloc error\n");
	    exit(1);
	}
	for (i=0; i<num_threads; i++) {
	    if ( i != id ) thread_list[count++] = i;
	}
	shuffle(thread_list,num_threads-1);	/* randomize the thread list */
    }

    do {
      int local_work_left;
      /* Continue until all local work has been exhausted. */
      do {
	while ( MCB_retrieve(CB,(clist_t *) CLIST[id]) ) {
	    spawn_clists(CLIST[id], id);
            if (id == 0) msg_handler_poll();
	}
        local_work_left = 0;	/* thread work queue is empty */
	if ( local_load_balancing ) {
          int i;
	  for (i=0; i<num_threads-1; i++) {
            MCB *target_cb = &mcb_list[thread_list[i]];
            int size = MCB_dequeue(target_cb, (clist_t *) CLIST[id], CB->compsub);
	    if (size != 0) {
              local_work_left = 1;	/* thread work queue non-empty */
	      rstats[id].local_fetch++;
	      rstats[id].local_bytes += (size*sizeof(clist_t));
	      spawn_clists(CLIST[id], id);
              if (id == 0) msg_handler_poll();
	      break;
	    }
	  }
	}
      } while (local_work_left);
      /* No local work left. */
      if ( remote_load_balancing ) {
            /* Try to get remote work. */
            double st = dclock();
            if (num_threads > 1) thread_barrier();	/* all threads idle */
            if (id == 0) {
              int i;
              /* Look for work for each thread. */
              global_work_left = 0;
              for (i=0;i < num_threads;i++) {
                int size = remote_dequeue((clist_t *) CLIST[i], mcb_list[i].compsub);
                new_work[i] = size;
                if (size != 0) {
	          rstats[i].remote_fetch++;
	          rstats[i].remote_bytes += (size*sizeof(clist_t));
                  global_work_left = 1;
                }
              }
            }
            if (num_threads > 1) thread_barrier();	/* all threads restart */
            rstats[id].fetch_time += dclock() - st;
	    if (new_work[id] != 0) spawn_clists(CLIST[id], id);
      }
    } while (global_work_left);

    thread_destroy();					/* free thread barrier */
    if (thread_list != NULL) free(thread_list);

    /* Save memory use and time statistics. */
    rstats[id].max_memory = CB->max_memory;
    rstats[id].enum_time += dclock() - enum_start;

    /* Dump the run time samples if saved during the run. */
    if (sample_time != 0) sample_write(id);

    /* Makes sure that the output buffer is flushed before returning */
    if ( OUTPUT_SET ) fflush(CG.CliqueFP[id]);

    return NULL;
}


#if defined(USE_PTHREADS)

static pthread_barrier_t lobby;		/* thread waiting area */

/* Initialize thread barrier (called by main thread only). */
static void thread_init(void)
{
  int errnum = pthread_barrier_init(&lobby,NULL,num_threads);
  if (errnum != 0) {
    fprintf(stderr,"pthread_barrier_init: %s",strerror(errnum));
    exit(1);
  }

  new_work = (int *) calloc(num_threads,sizeof(int));
  if (new_work == NULL) {
    fprintf(stderr,"%d: Unable to allocate new_work\n",mpi_rank);
    exit(1);
  }
}

/* Synchronize all threads (called by all threads). */
static void thread_barrier(void)
{
  int errnum = pthread_barrier_wait(&lobby);
  if (errnum > 0) {
    fprintf(stderr,"pthread_barrier_wait: %s",strerror(errnum));
    exit(1);
  }
}

/* Free thread barrier (called by all threads). */
static void thread_destroy(void)
{
  if (pthread_barrier_wait(&lobby) == PTHREAD_BARRIER_SERIAL_THREAD) {
     pthread_barrier_destroy(&lobby);
     if (new_work != NULL) free(new_work);
  }
}

#else

static void thread_init(void)    { }
static void thread_barrier(void) { }
static void thread_destroy(void) { }

#endif
