#include <stdio.h>
#include <stdlib.h>
#if defined(USE_MPI)
#include <mpi.h>
#endif
#include "graph.h"
#include "msg.h"

/* Size of an adjacency matrix word in bits. */
#define BIT_SIZE		(sizeof(graph_matrix_t) * 8)

/* Edge lookup hash table tunable values. */
static int min_hash_degree = 10;	/* minimum vertex degree to use the hash lookup */
static int hash_size_factor = 6;	/* multiple of vertex degree minimum table size */

/*
 * Allocate and initialize a graph, returns a pointer to it
 */
Graph *make_graph(graph_enum_t nvertices, graph_enum_t nedges)
{
    Graph *G;
    int base_size   = sizeof(Graph);		/* memory for base graph data */
    int vertex_size = nvertices*sizeof(Vertex);	/* memory for vertices */
    int edge_size   = 2*nedges*sizeof(Edge);	/* memory for edges */
    int edge_offset = base_size + vertex_size;	/* offset to start of edge memory */
    int graph_size  = edge_offset + edge_size;	/* total graph size */
    int i;

    G = (Graph *) malloc(graph_size);
    if (G == NULL) {
        fprintf(stderr,"Unable to allocate graph of size %d\n",graph_size);
        exit(1);
    }

    for (i=0; i<nvertices; i++) {
	G->vertex[i].degree = 0;
	G->vertex[i].active = 1;
	G->vertex[i].hash_list = NULL;	/* edge hash table */
	G->vertex[i].hash_size = 0;	/* size of the hash table */
    }

    G->num_edges           = nedges;
    G->num_vertices        = nvertices;
    G->num_active_vertices = nvertices;
    G->max_degree           = 0;		/* set in read_graph */
    G->graph_size           = graph_size;
    G->edge_offset          = edge_offset;	/* for use by remote processes */
    G->edge_array           = (Edge *) (((char *) G) + edge_offset);
    G->adjacency_matrix     = NULL;

    return G;
}


static void adj_matrix_add(Graph *G, graph_enum_t u, graph_enum_t v)
{
    int word_pos, bit_pos;
    graph_matrix_t bit_mask;

    /* Triangular matrix row is u and column is v, so make sure u > v. */
    if (u < v) {
      graph_enum_t t = u;
      u = v;
      v = t;
    }

    word_pos = v / BIT_SIZE;	/* index of matrix data word */
    bit_pos  = v % BIT_SIZE;	/* index of bit in word */
    bit_mask = ((graph_matrix_t) 1) << bit_pos;

    /* Set the bit in the matrix for connected(u,v). */
    G->adjacency_matrix[u][word_pos] |= bit_mask;
}


/*
 * Adjacency matrix connected test.
 * Return value is 1 if a (u,v) edge exists, 0 if (u,v) not connected.
 */
static int matrix_connected(Graph *G, graph_enum_t u, graph_enum_t v)
{
    int word_pos, bit_pos;
    graph_matrix_t bit_mask, word;

    /* Triangular matrix row is u and column is v, so make sure u > v. */
    if (u < v) {
      graph_enum_t t = u;
      u = v;
      v = t;
    }

    word_pos = v / BIT_SIZE;	/* index of matrix data word */
    bit_pos  = v % BIT_SIZE;	/* index of bit in word */
    bit_mask = (graph_matrix_t) 1 << bit_pos;
    word     = G->adjacency_matrix[u][word_pos];

    /* Check the bit in the matrix for connected(u,v). */
    return ((word & bit_mask) != 0);
}


/*
 * Linear list search of the adjacency list connected test.
 * Return value is 1 if a (u,v) edge exists, 0 if (u,v) not connected.
 */
static int list_connected(Graph *G, graph_enum_t u, graph_enum_t v)
{
    Vertex *vertex = &G->vertex[u];
    int size = vertex->degree;
    Edge *list = vertex->edge_list;

    for (int i=0;i < size;i++) {
        if (list[i].vertex == v) return 1;
    }

    return 0;
}


/*
 * Hash table search of the adjacency list connected test.
 * Return value is 1 if a (u,v) edge exists, 0 if (u,v) not connected.
 */
static int hash_connected(Graph *G, graph_enum_t u, graph_enum_t v)
{
    Vertex *vertex = &G->vertex[u];
    int i = (int) (v & vertex->hash_mask);	/* simple hash function */
    graph_enum_t *list = vertex->hash_list;	/* hash table of connected vertices */

    while (list[i] != -1) {
        if (list[i] == v) return 1;		/* found vertex */
        i++;					/* next edge */
        if (i == vertex->hash_size) i = 0;	/* wrap around */
    }

    return 0;
}


/*
 * Return true if vertex u is connected to vertex v.
 */
int connected(Graph *G, graph_enum_t u, graph_enum_t v)
{
  if ( u == v ) {
    return 1;	/* a vertex is connected to itself */
  } else if (G->adjacency_matrix != NULL) {
    return matrix_connected(G,u,v);
  } else if (G->vertex[u].hash_size != 0) {
    return hash_connected(G,u,v);
  } else {
    return list_connected(G,u,v);
  }
}


/*
 * Initialize for calls to connected().
 *
 * If the graph is small enough, then an adjacency triangular bit matrix is
 * allocated and initialized.
 */
static void adj_matrix_init(Graph *G)
{
  long n = G->num_vertices;	/* number of vertices in the graph */
  graph_enum_t u, v;
  graph_matrix_t *p;		/* memory block for triangular matrix */
  int i;
  long matrix_size = n * n;	/* number of entries in a square matrix. */

  /* Size of triangular bit matrix size in words (minimum size is 1 word per row). */
  long words = matrix_size / (BIT_SIZE * 2) + n;

  /* Allocate the matrix column of pointers to the row data. */
  G->adjacency_matrix = (graph_matrix_t **) malloc(n * sizeof(graph_matrix_t *));
  if (G->adjacency_matrix == NULL) {
    fprintf(stderr,"%d: Out of memory at adjacency matrix column\n",mpi_rank);
    exit(1);
  }

  /* Allocate space for the row data. */
  p = (graph_matrix_t *) calloc(words,sizeof(graph_matrix_t));
  if (p == NULL) {
      fprintf(stderr,"%d: Out of memory, adjacency matrix size %ld\n",mpi_rank,words*sizeof(graph_matrix_t));
      exit(1);
  }

  /* Set up the rows. */
  for (i=0;i < n;i++) {
    int word_size = i / BIT_SIZE + 1;		/* number of words for this row */
    G->adjacency_matrix[i] = p;
    p += word_size;
  }

  /* Populate the matrix. */
  for (u=0;u < n;u++) {
    Vertex *vertex = &G->vertex[u];
    int i;
    for (i=0;i < vertex->degree;i++) {
      v = vertex->edge_list[i].vertex;
      adj_matrix_add(G,u,v);
    }
  }

#if defined(MATRIX_CHECK)
  for (u=0;u < n;u++) {
    for (v=0;v < n;v++) {
      if ((u != v) && (matrix_connected(G,u,v) != list_connected(G,u,v))) {
        fprintf(stderr,"%d: Mismatch at (%d,%d): am=%d, ls=%d\n",mpi_rank,u,v,
          matrix_connected(G,u,v),list_connected(G,u,v));
        exit(1);
      }
    }
  }
#endif
}


/*
 * Add a vertex v to the edge hash table for this vertex.
 */
static void hash_table_add(Vertex *vertex, graph_enum_t v)
{
  int i = (int) (v & vertex->hash_mask);	/* simple hash function */
  graph_enum_t *list = vertex->hash_list;	/* hash table of connected vertices */

  /* Find an empty entry (in the presence of collisions). */
  while (list[i] != -1) {
    i++;				/* next edge */
    if (i == vertex->hash_size) i = 0;	/* wrap around */
  }

  /* Add the vertex to the table. */
  list[i] = v;
}


/*
 * Allocate and initialize the adjacency hash table for select vertices.
 */
static void hash_table_init(Graph *G)
{
  int hash_count = 0;		/* number of vertices using hash tables */
  size_t table_size = 0;	/* total hash table size for all vertices */
  graph_enum_t *tbl_ptr;	/* current location in table */

  /* First, determine the total table size. */
  for (graph_enum_t u=0;u < G->num_vertices;u++) {
    Vertex *vertex = &G->vertex[u];
    if (vertex->degree >= min_hash_degree) {
      int min_size = vertex->degree * hash_size_factor;
      graph_enum_t size = 1;	/* power of 2 table size */
      int log2_size = 0;	/* log base 2 of the table size */
      /* Determine the power of 2 hash table size. */
      while (size < min_size) {
        size = size << 1;
        log2_size++;
      }
      /* Need this check? assert(log2_size<sizeof(graph_enum_t)*8); */
      vertex->hash_size = size;
      vertex->hash_mask = size - 1;	/* size is a power of 2 */
      table_size += size;
    }
  }

  /* Allocate the table. */
  G->hash_table = (graph_enum_t *) malloc(table_size*sizeof(graph_enum_t));
  if (G->hash_table == NULL) {
    fprintf(stderr,"%d: Out of memory for adjacency hash table size %ld\n",mpi_rank,table_size);
    exit(1);
  }

  /* Initialize the table to all empty entries. */
  for (int i=0;i < table_size;i++) G->hash_table[i] = -1;

  /* Populate the table with the graph edges. */
  tbl_ptr = G->hash_table;		/* start at beginning of the table */
  for (graph_enum_t u=0;u < G->num_vertices;u++) {
    Vertex *vertex = &G->vertex[u];
    if (vertex->hash_size > 0) {
      vertex->hash_list = tbl_ptr;
      tbl_ptr += vertex->hash_size;
      /* Populate the table. */
      for (int i=0;i < vertex->degree;i++) {
        graph_enum_t v = vertex->edge_list[i].vertex;
        hash_table_add(vertex,v);
      }
      hash_count++;
    }
  }

  if (mpi_rank == 0) printf("Hash table of size %ld used for %d vertices, min degree %d, size factor %d\n",table_size,hash_count,min_hash_degree,hash_size_factor);
}


/*
 * Mark all vertices of degree lower than k as inactive.
 */
void delete_lower_degree_vertices(Graph *G, graph_enum_t k)
{
    int i;

    for (i=0; i<G->num_vertices; i++) {
	if ( G->vertex[i].degree < k-1 ) {
	    G->vertex[i].active = 0;
	    G->num_active_vertices--;
	}
    }
}


/*
 * Free the graph memory.
 */
void free_graph(Graph *G)
{
    if (G->hash_table != NULL) free(G->hash_table);

    if (G->adjacency_matrix != NULL) {
      free(G->adjacency_matrix[0]);	/* triangular matrix */
      free(G->adjacency_matrix);	/* row pointers */
    }

    free(G);
}


/*
 * Read in an unweighted edge-list file, return a pointer to the graph
 *
 * Only process rank 0 reads the graph file.
 * The memory for the graph is allocated in one large chunk to make it
 * easy to broadcast the graph to the remaining processes.
 */
Graph *read_graph(const char *filename, long matrix_cutoff, long hash_cutoff)
{
    graph_enum_t nvertices, nedges, nbuf[2];
    int reader_rank = 0;	/* process that reads the input file */
    FILE *fp;
    Graph *G;
    int i;

    /* Read the graph size parameters and distribute to all processes. */
    if (mpi_rank == reader_rank) {
        fp = fopen(filename,"r");
        if (fp == NULL) {
            fprintf(stderr,"Unable to open file \"%s\"\n",filename);
            exit(1);
        }
        fscanf(fp, "%d %d", &nvertices, &nedges);
        nbuf[0] = nvertices;
        nbuf[1] = nedges;
    }
#if defined(USE_MPI)
    if (mpi_size > 1) MPI_Bcast(nbuf,2,MPI_INT,reader_rank,MPI_COMM_WORLD);
#endif
    nvertices = nbuf[0];
    nedges    = nbuf[1];

    /* Allocate contiguous space for all the graph data. */
    G = make_graph(nvertices,nedges);

    if (mpi_rank == reader_rank) {
        graph_enum_t *edge_list;	/* array of edges (vertex pairs) */
        graph_enum_t *edge_count;	/* number of edges per vertex */
        int offset = 0;			/* edge_array offset */
        int ne = 0;			/* number of edges */

        /* Allocate a list of vertex pairs which define the edges. */
        edge_list  = (graph_enum_t *) malloc(2*nedges*sizeof(graph_enum_t));
        if (edge_list == NULL) {
            fprintf(stderr,"Unable to allocate edge_list of size %d\n",nedges);
            exit(1);
        }

        /* Allocate a counter of the number of edges for each vertex. */
        edge_count = (graph_enum_t *) calloc(nvertices,sizeof(graph_enum_t));
        if (edge_count == NULL) {
            fprintf(stderr,"Unable to allocate edge_count of size %d\n",nvertices);
            exit(1);
        }

        /* Read the edge data into the list and count arrays. */
        for (i = 0; i < nedges; i++) {
            graph_enum_t u, v;
 	    fscanf(fp, "%d", &u);
	    fscanf(fp, "%d", &v);
	    if (u >= nvertices) {
                fprintf(stderr,"Edge %d has bad vertex %d\n",i+1,u);
                continue;
            }
	    if (v >= nvertices) {
                fprintf(stderr,"Edge %d has bad vertex %d\n",i+1,v);
                continue;
            }
            /* Could test for (u>v) to eliminate redundant edges. */
            edge_list[ne++] = u;
            edge_list[ne++] = v;
            edge_count[u]++;
            edge_count[v]++;
        }

        fclose(fp);

        /* Use per-vertex edge counts to set the edge list pointer. */
        for (i=0;i < nvertices;i++) {
            G->vertex[i].edge_list = &G->edge_array[offset];
            G->vertex[i].active = 1;
            offset += edge_count[i];
            /* Track the maximum degree of any vertex. */
            if (edge_count[i] > G->max_degree) G->max_degree = edge_count[i];
        }
        if (offset != ne) fprintf(stderr,"Warning: Edges %d, offset %d\n",ne,offset);

        /* Copy the list of edges into the per-vertex lists of adjacent vertices. */
        for (i = 0; i < ne; i+=2) {
            graph_enum_t u = edge_list[i];
            graph_enum_t v = edge_list[i+1];
            G->vertex[u].edge_list[G->vertex[u].degree++].vertex = v;
            G->vertex[v].edge_list[G->vertex[v].degree++].vertex = u;
        }

        /* Free the work space. */
        free(edge_count);
        free(edge_list);
    }

#if defined(USE_MPI)
    /* Send the graph to all processes. */
    if (mpi_size > 1) MPI_Bcast(G,G->graph_size,MPI_BYTE,reader_rank,MPI_COMM_WORLD);
#endif

    if (mpi_rank != reader_rank) {
        /* Use per-vertex edge counts to set the edge list pointer. */
        int offset = 0;
        G->edge_array = (Edge *) (((char *) G) + G->edge_offset);
        for (i=0;i < nvertices;i++) {
            G->vertex[i].edge_list = &G->edge_array[offset];
            offset += G->vertex[i].degree;
        }
    }

    if (G->num_vertices <= matrix_cutoff) {
        /* Create adjacency matrix if graph is small enough. */
        adj_matrix_init(G);
    } else if (G->num_edges <= hash_cutoff) {
        /* Create edge hash table. */
        hash_table_init(G);
    }

    return G;
}


/*
 * Print the graph in the same format as the input file.
 */
void print_graph(Graph *G)
{
    graph_enum_t u;
    
    printf("%d\t%d\n",G->num_vertices,G->num_edges);
    for (u=0; u<G->num_vertices; u++) {
	Vertex *vertex = &G->vertex[u];
        graph_enum_t j;
        for (j=0;j < vertex->degree;j++) {
	    Edge *edge = &vertex->edge_list[j];
            graph_enum_t v = edge->vertex;
	    if (u < v) printf("%d\t%d\n",u,v);
	}
    }
}
