#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mm.h"

/* fp_size is the size of _sp relative to clist_t. */
/*unsigned int fp_size = sizeof(unsigned int)/sizeof(clist_t); */
unsigned int fp_size = 1;


#if defined(USE_PTHREADS)

#define lock_queue(CB) {\
    if ( pthread_mutex_lock(&CB->mutex) != 0 ) { \
      fprintf(stderr,"mutex lock: error\n"); \
      exit(1); \
    }\
}

#define unlock_queue(CB) {\
    if ( pthread_mutex_unlock(&CB->mutex) != 0 ) { \
      fprintf(stderr,"mutex lock: error\n"); \
      exit(1); \
    }\
}

#else

#define lock_queue(CB)
#define unlock_queue(CB)

#endif


/*
 * Initialize a memory control block for a graph with the number of nodes specified.
 */
void MCB_initialize(MCB *CB, int node_count)
{
    if ( (CB->compsub = (clist_t *)malloc(node_count*sizeof(clist_t))) == NULL ) {
	fprintf(stderr, "malloc: failed\n"); 
	exit(1);
    }

#if defined(USE_PTHREADS)
    /* Create mutex to control access to the MCB. */
    if ( pthread_mutex_init(&CB->mutex, NULL) != 0 ) {
	fprintf(stderr, "mutex initialization: failed\n"); 
	exit(1);
    }
#endif
    
    CB->clists = 0;
    CB->memory_level = 0;
    CB->max_memory = 0;
    CB->BTail = CB->BHead = NULL;
    MCB_expand(CB);
}


/**
 *
 * "MCB_expand" allocates a new Memory Block.
 *
 *
 */
void MCB_expand(MCB *CB)
{
    MB *L;

    if ( (L=malloc(sizeof(MB))) == NULL ) {
	fprintf(stderr, "malloc: failed\n"); 
	exit(1);
    }
  
    if ( (L->memory = malloc(BLOCK_SIZE*sizeof(clist_t))) == NULL ) {
	fprintf(stderr, "malloc: failed\n"); 
	exit(1);
    }

    L->_head = L->_sp = 0;
    L->_next = NULL;

    if ( CB->BTail == NULL ) {
        /* Insert first block. */
	CB->BHead = CB->BTail = L;
	CB->BTail->_prev = CB->BTail->_next = NULL;
    } else {
        /* Insert subsequent block. */
	CB->BTail->_next = L;
	L->_prev = CB->BTail;
	CB->BTail = L;
    }
}

/**
 *
 * "MCB_next" returns the start address of the current Memory Block where a new
 * clique list will be written.
 *
 * @param:
 *   size: maximumly possible memory size needed.
 *
 */
clist_t *MCB_next(MCB *CB, int size)
{
    /* If sufficient memory available in the current block, return it. */
    if ( (CB->BTail->_sp+size) < BLOCK_SIZE ) 
	return (CB->BTail->memory + CB->BTail->_sp);

    if ( CB->BTail->_next == NULL ) {
        /* Allocate a new block. */
	MCB_expand(CB);
    } else {
        /* Move to the next block in the list of blocks. */
	CB->BTail = CB->BTail->_next;
	CB->BTail->_head = CB->BTail->_sp = 0;
    }

    return CB->BTail->memory;
}


/**
 *
 * "MCB_advance" advances _sp pointer of the current MemoryBlock by
 * "size".  It is called when a clique list, which is appended at the
 * end of the current MemoryBlock, turns out to be a potential one.
 * 
 */
void MCB_advance(MCB *CB, int size)
{
    lock_queue(CB);

    /* Save current sp to start the frame. */
    *((unsigned int *)(CB->BTail->memory + CB->BTail->_sp + size )) = CB->BTail->_sp;

    /* Move the sp to the start of free memory at the end of the block. */
    CB->BTail->_sp += (size +  fp_size);

    /* Track total clique memory usage. */
    CB->memory_level += ((size+fp_size) * sizeof(clist_t));

    /* Track maximum clique memory usage. */
    if ( CB->memory_level > CB->max_memory ) CB->max_memory = CB->memory_level;

    /* Track the total number of clists for this MCB. */
    CB->clists++;

    unlock_queue(CB);
}


/**
 *
 * "MCB_retrieve" copies an unstructured CList in the current MemoryBlock
 * into a structured CList. It then adjusts "_sp" of the MemoryBlock.
 *
 */
int MCB_retrieve(MCB *CB, clist_t *L)
{
    int fp, size;

    /* Need to coordinate a possible race condition with a "stealer". */
    lock_queue(CB);

    /* Check whether the queue is empty */
    while ( CB->BTail->_sp == CB->BTail->_head ) {
        CB->BTail->_sp = CB->BTail->_head = 0;
	if ( CB->BTail == CB->BHead ) {
	  unlock_queue(CB);
	  return 0;
        }
	else {
	    CB->BTail = CB->BTail->_prev;
	}
    }

    /* Restore the last frame pointer */
    fp = *((unsigned int *)(CB->BTail->memory + CB->BTail->_sp - fp_size));

    size = (CB->BTail->_sp - fp - fp_size);
    CB->memory_level -= ((size+fp_size)*sizeof(clist_t));
    CB->BTail->_sp = fp;
    CB->clists--;

    memcpy(L,CB->BTail->memory+CB->BTail->_sp,size*sizeof(clist_t));

    unlock_queue(CB);
    return size;
}


/**
 *
 * "MCB_dequeue" takes off the first clist in the head MemoryBlock in the MCB.
 * It is used to steal a "workload" from other node.
 *
 * @param:
 *   CB: Memory Control Block from where a clist will be taken.
 *   block: buffer to copy a clist.
 *
 */
int MCB_dequeue(MCB *CB, clist_t *block, clist_t *compsub)
{
    clist_t depth;
    unsigned int size;
    cstruct_t *cs;

    lock_queue(CB);

    /* If the work queue is empty, return nothing. */
    if ( CB->BHead->_head == CB->BHead->_sp ) { 
      unlock_queue(CB);
      return 0;
    }

    cs = (cstruct_t *) (CB->BHead->memory + CB->BHead->_head);
    size = *(CB->BHead->memory + CB->BHead->_head + 3) + 4 + fp_size;

    /* If the work queue has only one item, return nothing. */
    if ( CB->BHead->_head + size == CB->BHead->_sp ) {
      unlock_queue(CB);
      return 0;
    }

    depth = *(CB->BHead->memory + CB->BHead->_head);

    /* Copy work out of the memory block into the caller arrays. */
    memcpy(compsub, CB->compsub, depth*sizeof(clist_t));
    memcpy(block, cs, size*sizeof(clist_t));

    /* Adjust MCB and memory block to remove the item. */
    CB->memory_level -= (size * sizeof(clist_t));
    CB->clists--;
    CB->BHead->_head += size;

    unlock_queue(CB);
 
    return (size+depth);
}
