/* TODO: allocate as maximum degree of the graph */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "clique.h"
#include "graph.h"
#include "stats.h"

int mpi_rank = 0;			/* rank of this MPI process */
int mpi_size = 1;			/* total number of MPI processes */
char *event_file = NULL;		/* output file name prefix for events */
static FILE *event_fp = NULL;		/* output file for events */
static double event_start = 0;		/* epoch start for events */

#if defined(USE_MPI)

#include <mpi.h>
#include "msg.h"

/* Message tags are used to create two virtual message channels, one
 * for requests, and one for responses. */
enum { request_tag, response_tag };

/* Message types. */
enum {
   dq_request=101,	/* request remote process dequeue */
   idle_notice,		/* notify that this process is going idle */
   dq_response,		/* response from a dequeue request */
   err_response,	/* response to an unknown request type */
};

/* Request message. */
typedef struct {
  int msg_type;		/* type of request */
  int from;		/* rank of request process */
  int dq_size;		/* number of items to dequeue (for dq_request) */
} request_t;

/* Generic (error) response. */
typedef struct {
  int msg_type;		/* type of response */
  int status;		/* -1 = error */
} response_t;

/* Dequeue request response. */
typedef struct {
  int msg_type;		/* type of response */
  int status;		/* number of items dequeued or -1 for error */
  clist_t clist[];	/* num items of clist data followed by compsub data */
} dq_response_t;

static int dequeue_check( int dest, clist_t *clist, clist_t *compsub);
static void process_request_messages(void);
static void msg_handler_post(void);
static void msg_handler_recv(void);
static void dequeue_request_message(int dest, int dq_size);
static void idle_notice_message(int dest);
static void unknown_request_message(int dest);

static int debug = 0;			/* debug print flag */

static int request_total = 0;		/* number of requests received */
static int idle_notice_total = 0;	/* number of idle notices received */

static MPI_Comm comm = MPI_COMM_WORLD;	/* communicator to use */
static MPI_Request msg_request;		/* for incoming request messages */
static request_t posted_req;		/* incoming request buffer */
static dq_response_t *dq_recv_buf;	/* dequeue response message receive buffer */
static dq_response_t *dq_send_buf;	/* dequeue response message send buffer */
static int response_size;		/* size of dequeue response buffer */

/*
 * This data is used in handling dequeue requests from other processes.
 */
static clist_t *clist, *compsub;	/* buffers for local dequeue */
static int *local_target;		/* random list of local threads */
static int *remote_target;		/* random list of remote processes */
static int *idle;			/* processes which have sent idle notices */


/*
 * Search all remote processes until one returns some work.
 *
 * Return value is 0 if no process has work to share.  Otherwise,
 * return value is the total size of the items in clist and compsub.
 */
int remote_dequeue(clist_t *clist, clist_t *compsub)
{
    int i;
    int status = 0;
    int count = 0;
    double st = dclock();

    /* Get a random list of processes (other than this one). */
    shuffle(remote_target,mpi_size-1);

    /* Try each process in the list until one shares some work. */
    for (i=0;i < mpi_size-1;i++) {
        status = dequeue_check(remote_target[i],clist,compsub);
        count++;
        if (status != 0) break;	/* found work */
    }

    if (event_fp != NULL) {
      double time = dclock() - st;
      double when = st - event_start;
      fprintf(event_fp,"%-6s %8.3f %8.6f %5d\n","fetch",when,time,count);
    }

    return status;
}


/*
 * Poll for request messages (with thread locking).
 */
void msg_handler_poll(void)
{
  process_request_messages();
}


int msg_handler_init(int *pargc, char ***pargv)
{
    if (MPI_Init(pargc,pargv) != MPI_SUCCESS) {
      fprintf(stderr,"pDFS: Unable to initialize MPI.\n");
      exit(1);
    }

    MPI_Comm_size(comm,&mpi_size);
    MPI_Comm_rank(comm,&mpi_rank);

    return mpi_rank;
}


/*
 * After the command line options are processed and the graph read,
 * get ready and then begin processing request messages.
 * dq_max is the largest number if items dequeued in a single request.
 */
void msg_handler_start(int dq_max)
{
    int n = 0;
    int i;
    int max_clist_size = G->num_vertices * sizeof(clist_t);

    /* This array is used to hold a clist stolen from a local thread. */
    clist = (clist_t *) malloc(max_clist_size+4);
    if (clist == NULL ) {
	fprintf(stderr,"Unable to allocate clist in message handler.\n");
	exit(1);
    }

    /* This array is used to hold a compsub stolen from a local thread. */
    compsub = (clist_t *) malloc(max_clist_size+4);
    if (compsub == NULL ) {
	fprintf(stderr,"Unable to allocate compsub in message handler.\n");
	exit(1);
    }

    /* This buffer is used to hold a dq_response received from a process. */
    response_size = sizeof(dq_response_t) + (G->max_degree+8) * sizeof(clist_t) * dq_max;
    dq_recv_buf = (dq_response_t *) malloc(response_size);
    if (dq_recv_buf == NULL ) {
	fprintf(stderr,"Unable to allocate dequeue response buffer in message handler.\n");
	exit(1);
    }

    /* This buffer is used to hold a dq_response sent to a process. */
    dq_send_buf = (dq_response_t *) malloc(response_size);
    if (dq_send_buf == NULL ) {
	fprintf(stderr,"Unable to allocate dequeue response buffer in message handler.\n");
	exit(1);
    }

    /* This array is a randomly order list of local threads for remote_dequeue. */
    local_target = (int *) malloc(num_threads*sizeof(int));
    if (local_target == NULL ) {
	fprintf(stderr,"Unable to allocate local_target in message handler.\n");
	exit(1);
    }

    /* This array is a randomly order list of remote processes. */
    remote_target = (int *) malloc(mpi_size*sizeof(int));
    if (remote_target == NULL ) {
	fprintf(stderr,"Unable to allocate remote_target in message handler.\n");
	exit(1);
    }

    /* This array of process flags reduces the number of remote dequeue requests. */
    idle = (int *) calloc(mpi_size,sizeof(int));
    if (idle == NULL ) {
	fprintf(stderr,"Unable to allocate idle in message handler.\n");
	exit(1);
    }

    /* Initialize the list of local threads (all threads). */
    for (i=0;i < num_threads;i++) local_target[i] = i;

    /* Initialize the list of remote processes (all except this process). */
    for (i=0;i < mpi_size;i++) if (i != mpi_rank) remote_target[n++] = i;

    /* If an event file prefix was specified, open the event file. */
    if (event_file != NULL) {
      char fname[1024];
      sprintf(fname,"%s.%04d",event_file,mpi_rank);
      event_fp = fopen(fname,"w");
      if (event_fp == NULL) {
        fprintf(stderr,"Unable to open event file \"%s\"\n",fname);
        exit(1);
      }
      fprintf(event_fp,"%-6s %-8s %-8s %-5s\n","Event","When","Time","Count");
      event_start = dclock();
    }

    /* Post the first buffer for a request. */
    msg_handler_post();
}


/*
 * When this process is idle (no local thread can share work and no remote
 * process can share work), then get ready to shut down.
 */
void msg_handler_complete(void)
{
    int dest;
    request_t req;
    double st = dclock();

    /* Send out idle notices to all other processes. */
    req.msg_type = idle_notice;
    req.from     = mpi_rank;
    for (dest=0; dest < mpi_size;dest++) {
       if (dest == mpi_rank) continue;	/* don't send to self */
       if (debug) fprintf(stderr,"%d: Sending idle notice %d to %d\n",mpi_rank,idle_notice,dest);
       if (MPI_Send(&req,sizeof(request_t),MPI_BYTE,dest,0,comm) != MPI_SUCCESS) {
           fprintf(stderr,"%d: Unable to send a response to %d\n",mpi_rank,dest);
       }
       /* Continue to process requests while shutting down. */
       process_request_messages();
    }

    /* Wait for all other processes to check in with idle notices. */
    if (debug) fprintf(stderr,"%d: idle wait, total is %d, expecting %d\n",mpi_rank,idle_notice_total,mpi_size-1);
    while (idle_notice_total < mpi_size-1) {
       process_request_messages();
    }

    if (event_fp != NULL) {
      double time = dclock() - st;
      double when = st - event_start;
      fprintf(event_fp,"%-6s %8.3f %8.6f\n","idle",when,time);
      fclose(event_fp);
    }

    free(remote_target);
    free(local_target);
    free(idle);
    free(dq_recv_buf);
    free(dq_send_buf);
    free(compsub);
    free(clist);
}


void msg_handler_stop(void)
{
    if (MPI_Cancel(&msg_request) != MPI_SUCCESS) {
        fprintf(stderr, "%d: Unable to cancel pending communication.\n", mpi_rank);
    }

    if (debug) fprintf(stderr,"%d: Finalize\n",mpi_rank);
    if (MPI_Finalize() != MPI_SUCCESS) {
	fprintf(stderr,"%d: Unable to finalize MPI.\n",mpi_rank);
    }
}


/*
 * Attempt to get work from another process.
 *
 * Return value is 0 if the process has no work to share.  Otherwise,
 * return value is the total size of the items in clist and compsub.
 */
static int dequeue_check( int dest, clist_t *clist, clist_t *compsub)
{
    request_t req;
    MPI_Status status;
    MPI_Request dequeue_req;
    int msg_size;
    int size = 0;
    int depth = 0;
    int flag = 0;

    /* If the dest process has already sent the idle notice, it has no work. */
    if (idle[dest]) return 0;

    req.msg_type = dq_request;
    req.from     = mpi_rank;
    req.dq_size  = 1;

    /* Post the response buffer. */
    if (MPI_Irecv(dq_recv_buf,response_size,MPI_BYTE,dest,response_tag,comm,&dequeue_req) != MPI_SUCCESS) {
        fprintf(stderr,"%d: Unable to receive a response from %d\n",mpi_rank,dest);
        MPI_Abort(MPI_COMM_WORLD,1);
    }

    if (debug) fprintf(stderr,"%d: Sending dq_request msg_type %d to %d\n",mpi_rank,req.msg_type,dest);
    if (MPI_Send(&req,sizeof(request_t),MPI_BYTE,dest,request_tag,comm) != MPI_SUCCESS) {
        fprintf(stderr,"%d: Unable to send a request to %d\n",mpi_rank,dest);
        MPI_Abort(MPI_COMM_WORLD,1);
    }

    /* Wait for response to the request.  Poll both response and incoming requests. */
    for (;;) {
        /* Poll for response. */
        if (MPI_Test(&dequeue_req,&flag,MPI_STATUS_IGNORE) != MPI_SUCCESS) {
            fprintf(stderr,"%d: Unable to get status of response from %d\n",mpi_rank,dest);
            MPI_Abort(MPI_COMM_WORLD,1);
        }
        if (flag) break;		/* response received */
        /* Poll for incoming requests. */
        process_request_messages();
    }

    if (debug) fprintf(stderr,"%d: Receiving a response from %d\n",mpi_rank,dest);
    memset(&status,0,sizeof(MPI_Status));
    if (MPI_Wait(&dequeue_req,&status) != MPI_SUCCESS) {
        fprintf(stderr,"%d: Unable to wait for a dequeue response.\n",mpi_rank);
    } else {
      if (debug) fprintf(stderr,"%d: Response type from %d is %d, status %d\n",mpi_rank,dest,dq_recv_buf->msg_type,dq_recv_buf->status);
      if (dq_recv_buf->status > 0) {
          size  = dq_recv_buf->clist[3] + 3 + 1;
          depth = dq_recv_buf->clist[0];
          memcpy(clist,&dq_recv_buf->clist[0],size * sizeof(clist_t));
          memcpy(compsub,&dq_recv_buf->clist[size],depth * sizeof(clist_t));
      }
    }

    return size + depth;
}


/*
 * Process outstanding request messages.
 * */
static void process_request_messages(void)
{
  int flag = 0;

  do {
      if (0) fprintf(stderr,"%d: Request poll.\n",mpi_rank);
      if (MPI_Test(&msg_request,&flag,MPI_STATUS_IGNORE) != MPI_SUCCESS) {
        fprintf(stderr,"%d: Unable to poll for request.\n",mpi_rank);
      }
      if (flag) {
          /* Process an outstanding request message. */
          msg_handler_recv();
          /* Post the buffer for the next request. */
          msg_handler_post();
      }
  } while (flag);
}


/*
 * Post a buffer for a non-blocking receive using the request tag.
 */
static void msg_handler_post(void)
{
  int src = MPI_ANY_SOURCE;

  if (0) fprintf(stderr,"%d: Request post.\n",mpi_rank);
  if (MPI_Irecv(&posted_req,sizeof(request_t),MPI_BYTE,src,request_tag,comm,&msg_request) != MPI_SUCCESS) {
      fprintf(stderr,"%d: Unable to post a request buffer.\n",mpi_rank);
  }
}


/*
 * Receive and process a request message.
 */
static void msg_handler_recv(void)
{
    MPI_Status status;
    int dest;			/* response destination process rank */
    double st = dclock();

    memset(&status,0,sizeof(MPI_Status));
    if (MPI_Wait(&msg_request,&status) != MPI_SUCCESS) {
        fprintf(stderr,"%d: Unable to wait for a request.\n",mpi_rank);
        return;
    }
    request_total++;
    if (status.MPI_ERROR != MPI_SUCCESS) {
        /* Consider this non-fatal. */
        if (debug) fprintf(stderr,"%d: Message error %d received while waiting for a request.\n",mpi_rank,status.MPI_ERROR);
    }
    dest = posted_req.from;
    if (debug) fprintf(stderr,"%d: Received request type %d from %d\n",mpi_rank,posted_req.msg_type,dest);

    /* Handle the request message. */
    if (posted_req.msg_type == dq_request) {
        dequeue_request_message(dest,posted_req.dq_size);
    } else if (posted_req.msg_type == idle_notice) {
        idle_notice_message(dest);
    } else {
        unknown_request_message(dest);
    }

    rstats[0].request_time += dclock() - st;
}


/*
 * Handle a dequeue request message from another process.
 */
static void dequeue_request_message(int dest, int dq_size)
{
    int msg_size;		/* size of the dequeue response message */
    int empty_count = 0;	/* number of empty queues in a row */
    int idx = 0;		/* target thread index */
    int num_items = 0;		/* number of items successfully dequeued */
    int p = 0;			/* response message clist index */

    /* Look for dq_size items of work to steal from any local thread.
     * Uses a random sequence of queues via local_target, stealing an
     * equal number of items from all queues with available work to share. */
    shuffle(local_target,num_threads);
    while ((empty_count < num_threads) && (num_items < dq_size)) {
      MCB *target_cb = &mcb_list[local_target[idx]];
      int status;		/* local dequeue status */
      /* Check for available space before attempting dequeue. */
      if (sizeof(response_t)+(p+G->max_degree+8)*sizeof(clist_t) > response_size) {
        fprintf(stderr,"%d: Invalid dequeue request for %d items.\n",mpi_rank,dq_size);
        break;
      }
      /* See if this queue has any work to share. */
      status = MCB_dequeue(target_cb,clist,compsub);
      if (status == 0) {
        empty_count++;		/* no items available from this queue */
      } else {
        int size  = *(clist+3) + 3 + 1;
        int depth = *clist;
        memcpy(&dq_send_buf->clist[p],clist,size * sizeof(clist_t));
        p += size;		/* move pointer to end of data */
        memcpy(&dq_send_buf->clist[p],compsub,depth * sizeof(clist_t));
        p += depth;		/* move pointer to end of data */
        empty_count = 0;
        num_items++;
      }
      /* Move to the next queue. */
      if (++idx == num_threads) idx = 0;
    }

    /* Initialize the dequeue response message. */
    msg_size = sizeof(response_t) + p*sizeof(clist_t);
    dq_send_buf->msg_type = dq_response;
    dq_send_buf->status = num_items;

    /* Send the message. */
    if (debug) fprintf(stderr,"%d: Sending response msg_type %d of size %d to %d, status %d\n",mpi_rank,dq_send_buf->msg_type,msg_size,dest,dq_send_buf->status);
    if (MPI_Send(dq_send_buf,msg_size,MPI_BYTE,dest,response_tag,comm) != MPI_SUCCESS) {
        fprintf(stderr,"%d: Unable to send a response to %d\n",mpi_rank,dest);
    }
}


/*
 * Record an idle process.
 */
static void idle_notice_message(int dest)
{
    if (debug) fprintf(stderr,"%d: Received idle notice from %d\n",mpi_rank,dest);
    idle[dest] = 1;
    idle_notice_total++;
    if (debug) fprintf(stderr,"%d: idle notice total is %d, expecting %d\n",mpi_rank,idle_notice_total,mpi_size-1);
}


/*
 * Request type not recognized.  Return error status.
 * This should never happen, so maybe an abort would be a better solution.
 */
static void unknown_request_message(int dest)
{
    response_t r;

    r.msg_type = err_response;
    r.status = -1;
    if (debug) fprintf(stderr,"%d: Sending error response to %d, status %d\n",mpi_rank,dest,r.status);
    if (MPI_Send(&r,sizeof(response_t),MPI_BYTE,dest,response_tag,comm) != MPI_SUCCESS) {
        fprintf(stderr,"%d: Unable to send a response to %d\n",mpi_rank,dest);
    }
}


#else

/*
 * If not using MPI, just return 0 (no work to share).
 */
int remote_dequeue(clist_t *clist, clist_t *compsub) { return 0; }

/*
 * msg_handler stubs for non-MPI version.
 */
int msg_handler_init(int *pargc, char ***pargv) { return 0; };
void msg_handler_start(void) { }
void msg_handler_poll(void) { }
void msg_handler_complete(void) { }
void msg_handler_stop(void) { }

#endif
