#include <stdio.h>
#include "mm.h"
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#if defined(USE_MPI)
#include <mpi.h>
#endif

#include "stats.h"
#include "clique.h"
#include "msg.h"

/* Number of double-sized values in the rstats. */
#define NSTATS	sizeof(rstats_t) / sizeof(double)

rstats_t *rstats = NULL;		/* per-thread statistics */
double sample_time = 0;			/* sampling interval (0 = disable) */
int clique_breakdown = 0;		/* print clique counts by size */

/* Time-varying statistics. */
typedef struct {
  double time_stamp;		/* time of the sample */
  long clists;			/* number of items on the queue */
  long memory;			/* size of the queue in bytes */
} rsamp_t;

static rsamp_t *rsamp_list = NULL;	/* run-time samples */
static int rsamp_count = 0;		/* number of items on the list */
static int rsamp_size = 0;		/* size of the run-time sample list */
static double last_sample = 0;		/* time of the last sample */
static long *clique_sizes = NULL;	/* clique size table */


/*
 * Check to see if a run time statistics sample needs to be saved.
 */
void sample_check(double current_time, int id)
{
  double tick = current_time - last_sample;
  MCB *CB = &mcb_list[id];

  if (tick < sample_time) return;	/* not time yet */

  last_sample = current_time;

  /* If no space for a sample, expand the list. */
  if (rsamp_count >= rsamp_size) {
	rsamp_size += 1000;
	rsamp_list = (rsamp_t *) realloc(rsamp_list,rsamp_size*sizeof(rsamp_t));
	if (rsamp_list == NULL) {
	    fprintf(stderr,"%d: Not enough memory for %d samples!\n",mpi_rank,rsamp_size);
	    rsamp_size = 0;
	    rsamp_count = 0;
	}
  }

  /* If space is available, add a sample. */
  if (rsamp_count < rsamp_size) {
	rsamp_list[rsamp_count].time_stamp = current_time;
	rsamp_list[rsamp_count].clists = CB->clists;
	rsamp_list[rsamp_count].memory = CB->memory_level;
	rsamp_count++;
  }
}


/*
 * Write the run time samples to a file.
 */
void sample_write(int id)
{
  FILE *fp;
  char fname[20];

  if (rsamp_count == 0) return;	/* no samples */

#if defined(USE_MPI) && defined(USE_THREADS)
  sprintf(fname,"sample.%04d.%04d",mpi_rank,id);
#elif defined(USE_MPI)
  sprintf(fname,"sample.%04d",mpi_rank);
#elif defined(USE_THREADS)
  sprintf(fname,"sample.%04d",id);
#else
  strcpy(fname,"sample.out");
#endif

  fp = fopen(fname,"w");
  if (fp == NULL) {
    fprintf(stderr,"%d: Unable to open file \"%s\"\n",mpi_rank,fname);
  } else {
    int i;
    double start_time = rsamp_list[0].time_stamp;
    fprintf(fp,"%d samples at %.2f second intervals for process rank %d, thread %d\n",
      rsamp_count,sample_time,mpi_rank,id);
    fprintf(fp,"%-8s %-8s %-8s\n","Time","Clists","Memory");
    for (i=0;i < rsamp_count;i++) {
      double t= rsamp_list[i].time_stamp - start_time;
      fprintf(fp,"%8.2f %8d %8d\n",t,rsamp_list[i].clists,rsamp_list[i].memory);
    }
    fclose(fp);
  }
}


static void print_rstats(const rstats_t *local, int nlocal)
{
  int gcount = nlocal * mpi_size;	/* global thread count */
  int dcount = NSTATS * nlocal;
  rstats_t *r;
  
#if defined(USE_MPI)
  r = (rstats_t *) malloc(gcount*sizeof(rstats_t));
  if (r == NULL) {
    fprintf(stderr,"Unable to allocate %d rstats\n",gcount);
    exit(1);
  }

  MPI_Gather((double *) local,dcount,MPI_DOUBLE,r,dcount,MPI_DOUBLE,0,MPI_COMM_WORLD);
#else
  r = local;
#endif

  if (mpi_rank == 0) {
    int i, j;
    printf("\nProc Thr Total cliques\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12ld\n",i,j,(long) r[i*nlocal+j].cliques);
    printf("\nProc Thr Maximum memory used\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12ld\n",i,j,(long) r[i*nlocal+j].max_memory);
#if defined(USE_MPI)
    printf("\nProc Thr Remote fetches\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12ld\n",i,j,(long) r[i*nlocal+j].remote_fetch);
    printf("\nProc Thr Remote bytes\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12ld\n",i,j,(long) r[i*nlocal+j].remote_bytes);
#endif
#if defined(USE_PTHREADS)
    printf("\nProc Thr Local fetches\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12ld\n",i,j,(long) r[i*nlocal+j].local_fetch);
    printf("\nProc Thr Local bytes\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12ld\n",i,j,(long) r[i*nlocal+j].local_bytes);
#endif
#if defined(USE_MPI)
    printf("\nProc Thr Fetch time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].fetch_time);
    printf("\nProc Thr Proc Thr Request time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].request_time);
    printf("\nProc Thr Idle time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].idle_time);
#endif
    printf("\nProc Thr Init time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].init_time);
    printf("\nProc Thr Enumeration time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].enum_time);
    printf("\nProc Thr Compute time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].compute_time);
    printf("\nProc Thr Read time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].read_time);
    printf("\nProc Thr Root time\n");
    for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
      printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].root_time);
    if (OUTPUT_SET) {
      printf("\nProc Thr Clique output time\n");
      for (i=0;i < mpi_size;i++) for (j=0;j < nlocal;j++)
        printf("%4d %3d %12.3f\n",i,j,r[i*nlocal+j].output_time);
    }
    fputc('\n',stdout);
  }

#if defined(USE_MPI)
  free(r);
#endif
}


/*
 * Print global max, min, and average statistics.
 */
static void print_gstat(const char *label, double total, double min, double max, int integer)
{
    int gthreads = num_threads * mpi_size;
    double avg = total / gthreads;
    double imb = (max == 0) ? 0 : (max - min) / max * 100.0;
    if (integer) {
      printf("%-14s %12ld %12ld %12ld %12.1f%%\n",label,(long) max,(long) min,(long) avg,imb);
    } else {
      printf("%-14s %12.3f %12.3f %12.3f %12.1f%%\n",label,max,min,avg,imb);
    }
}


/*
 * Sum the individual statistics for all local threads.
 */
static void local_sum(rstats_t *r_in, int nstats, rstats_t *r_out)
{
  double *in  = (double *) r_in;
  double *out = (double *) r_out;
  int i, j;

  for (j=0;j < NSTATS;j++) out[j] = in[j];

  for (i=1;i < nstats;i++) {
    for (j=0;j < NSTATS;j++) {
      out[j] += in[i*NSTATS+j];
    }
  }
}


/*
 * Find the maximum of the individual statistics for all local threads.
 */
static void local_max(rstats_t *r_in, int nstats, rstats_t *r_out)
{
  double *in  = (double *) r_in;
  double *out = (double *) r_out;
  int i, j;

  for (j=0;j < NSTATS;j++) out[j] = in[j];

  for (i=1;i < nstats;i++) {
    for (j=0;j < NSTATS;j++) {
      double v = in[i*NSTATS+j];
      if (v > out[j]) out[j] = v;
    }
  }
}


/*
 * Find the minimum of the individual statistics for all local threads.
 */
static void local_min(rstats_t *r_in, int nstats, rstats_t *r_out)
{
  double *in  = (double *) r_in;
  double *out = (double *) r_out;
  int i, j;

  for (j=0;j < NSTATS;j++) out[j] = in[j];

  for (i=1;i < nstats;i++) {
    for (j=0;j < NSTATS;j++) {
      double v = in[i*NSTATS+j];
      if (v < out[j]) out[j] = v;
    }
  }
}


void rstats_init(void)
{
  if ( (rstats = (rstats_t *)calloc(num_threads, sizeof(rstats_t))) == NULL ) {
      fprintf(stderr,"rstats malloc error\n");
      exit(1);
  }
}


void rstats_end(void)
{
    rstats_t lmax, lmin, lsum;	/* local  min/max/sum */
    rstats_t gmax, gmin, gsum;	/* global min/sum/sum */
    int i;

    if (clique_breakdown) {
      long *size_sum = (long *) calloc(G->max_degree,sizeof(long));
      if (size_sum == NULL) {
        fprintf(stderr,"Out of memory.\n");
        exit(1);
      }
#if defined(USE_MPI)
      MPI_Reduce(clique_sizes,size_sum,G->max_degree,MPI_LONG,MPI_SUM,0,MPI_COMM_WORLD);
#else
      memcpy(size_sum,clique_sizes,G->max_degree*sizeof(long));
#endif
      if (mpi_rank == 0) {
        fputs("Clique counts by size\n",stdout);
        for (i=0;i < G->max_degree;i++) {
          if (size_sum[i] != 0) printf("%d: %ld\n",i,size_sum[i]);
        }
        fputc('\n',stdout);
      }
    }

    if (full_statistics) print_rstats(rstats,num_threads);

    /* Get local thread values. */
    local_sum(rstats,num_threads,&lsum);
    local_min(rstats,num_threads,&lmin);
    local_max(rstats,num_threads,&lmax);

    free(rstats);

    /* Get global statistics. */
#if defined(USE_MPI)
    MPI_Reduce(&lsum,&gsum,NSTATS,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce(&lmax,&gmax,NSTATS,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
    MPI_Reduce(&lmin,&gmin,NSTATS,MPI_DOUBLE,MPI_MIN,0,MPI_COMM_WORLD);
#else
    gsum = lsum;
    gmax = lmax;
    gmin = lmin;
#endif

    if (mpi_rank == 0) {
        printf("%ld cliques in %.3f seconds, %.3f cliques/second/process.\n\n",
   	  (long) gsum.cliques,gsum.enum_time,gsum.cliques/gsum.enum_time);

        printf("%-14s %12s %12s %12s %12s\n","Performance","Maximum","Minimum","Average","Balance");
        print_gstat("Cliques",gsum.cliques,gmin.cliques,gmax.cliques,1);
        print_gstat("Memory",gsum.max_memory,gmin.max_memory,gmax.max_memory,1);
#if defined(USE_MPI)
        print_gstat("Remote fetch",gsum.remote_fetch,gmin.remote_fetch,gmax.remote_fetch,1);
        print_gstat("Remote bytes",gsum.remote_bytes,gmin.remote_bytes,gmax.remote_bytes,1);
#endif
#if defined(USE_PTHREADS)
        print_gstat("Local fetch",gsum.local_fetch,gmin.local_fetch,gmax.local_fetch,1);
        print_gstat("Local bytes",gsum.local_bytes,gmin.local_bytes,gmax.local_bytes,1);
#endif
        print_gstat("Init time",gsum.init_time,gmin.init_time,gmax.init_time,0);
        print_gstat("Read time",gsum.read_time,gmin.read_time,gmax.read_time,0);
        print_gstat("Root time",gsum.root_time,gmin.root_time,gmax.root_time,0);
        if (OUTPUT_SET) print_gstat("Output time",gsum.output_time,gmin.output_time,gmax.output_time,0);
        print_gstat("Enumerate time",gsum.enum_time,gmin.enum_time,gmax.enum_time,0);
        print_gstat("Compute time",gsum.compute_time,gmin.compute_time,gmax.compute_time,0);
        print_gstat("Fetch time",gsum.fetch_time,gmin.fetch_time,gmax.fetch_time,0);
        print_gstat("Request time",gsum.request_time,gmin.request_time,gmax.request_time,0);
        print_gstat("Idle time",gsum.idle_time,gmin.idle_time,gmax.idle_time,0);
    }
}


/*
 * Track the number of cliques found by size.
 */
void clique_bin(int size)
{
  /* Need to allocate here because rstats_init is called too early. */
  if (clique_sizes == NULL) {
    clique_sizes = (long *) calloc(G->max_degree,sizeof(long));
    if (clique_sizes == NULL) {
      fprintf(stderr,"rstats malloc error\n");
      exit(1);
    }
  }

  clique_sizes[size-1]++; /* Note: not per thread */
}


#if !defined(__QK_USER__)

double dclock(void)
{
  struct timeval tv;
  gettimeofday(&tv,0);
  return (double) tv.tv_sec + (double) tv.tv_usec * 1e-6;
}

#endif
